
#ifndef __DD_BIT_PATTERN_TREE_CACHE_ENTRY__
#define __DD_BIT_PATTERN_TREE_CACHE_ENTRY__

#include <vector>
#include <sstream>
#include <algorithm>
#include <math.h>

#include "DDStats.h"

#include "DDOptions.h"

#include "DDBitList.h"
#include "DDBitCountTable.h"
#include "DDSingleSetRay.h"
#include "DDBitSelector.h"

#include "DDMostCommonBitFirstStrategy.h"
#include "DDMostAverageBitFirstStrategy.h"
#include "DDLeastAverageBitFirstStrategy.h"
#include "DDLeastCommonBitFirstStrategy.h"
#include "DDFirstAvailableBitStrategy.h"
#include "DDNarrowQueryBitsNeutralizationStrategy.h"
#include "DDVerQueryBitsNeutralizationStrategy.h"

#include "DDIntermediateBitSetNode.h"
#include "DDLeafNode.h"
#include "DDVerificationCondition.h"
/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "DDExtVerificationCondition.h"

#include "DDRayReferenceContainer.h"

#include "DDBitSplitVisitor.h"
#include "DDSingleBitSplitVisitor.h"

#include "DDBinaryStaticVantagePointVisitor.h"
#include "DDBinaryDynamicVantagePointVisitor.h"

#include "DDPreliminaryTraverseVisitor.h"
#include "DDVerificationTraverseVisitor.h"
#include "DDBitFrequencyModifier.h"
#include "DDZeroFrequencyModifier.h"

#include "DDDegeneracyCalcVisitor.h"
#include "DDRayRefReleaseVisitor.h"

#include "DDPreliminaryCondition.h"
#include "DDVantagePointPreliminaryCondition.h"

namespace double_description {
	
	template <class RAY_TYPE>
	class BitPatternTreeCacheEntry {

		private:
        const unsigned short mRows;
		unsigned short mPlane;
		unsigned short mBlocks;
        unsigned int mMinCommonBits;

		RayReferenceContainer mPositiveDegEntries;
        RayReferenceContainer mPositiveNonDegEntries;
		RayReferenceContainer mZeroEntries;
        RayReferenceContainer mNegativeEntries;

        BitList mCutMap;

		public:
		BitPatternTreeCacheEntry ( const unsigned short rows,
                                   const unsigned short columns,
                                   const unsigned short current ) :
            mRows( rows ),
			mPlane( current ),
			mBlocks( BitList::getBlocks( rows ) ),
            mMinCommonBits( columns - 2 ),
            mCutMap( mBlocks, 0 ) {

                mCutMap.setBits( 0, mPlane );
            }

		BitPatternTreeCacheEntry ( const BitPatternTreeCacheEntry & other ) :
            mRows( other.mRows ),
			mPlane( other.mPlane ),
			mBlocks( other.mBlocks ),
            mMinCommonBits( other.mMinCommonBits ),
            mCutMap( other.mCutMap ) {}

        ~BitPatternTreeCacheEntry () {
            const RayRefReleaseVisitor<RAY_TYPE> release_visitor;

            mNegativeEntries.modify( release_visitor );
        }

		void insertPositive ( const RAY_TYPE* const ray ) {
            const RayReference ref( static_cast<const void*>( ray ), ray->getActiveSet() & mCutMap );
            
            if ( ref.activeSet.getPopulationCount() == mMinCommonBits + 1 ) {
                mPositiveNonDegEntries.insert( ref );
            } else {    
			    mPositiveDegEntries.insert( ref );
                mZeroEntries.insert( ref );
            }
		}

		void insertNegative ( const RAY_TYPE* const ray ) {
            const RayReference ref( static_cast<const void*>( ray ), ray->getActiveSet() & mCutMap );

			mNegativeEntries.insert( ref );

            if ( ref.activeSet.getPopulationCount() > mMinCommonBits + 1 ) {
                mZeroEntries.insert( ref );
            }
		}

		void insertZero ( const RAY_TYPE* const ray ) {
            const RayReference ref( static_cast<const void*>( ray ), ray->getActiveSet() & mCutMap );
	
			mZeroEntries.insert( ref );
		}

		unsigned int getPositiveCount () const {
			return mPositiveDegEntries.size() + mPositiveNonDegEntries.size();
		}

		unsigned int getNegativeCount () const {
			return mNegativeEntries.size();
		}

		void clear() {
            const RayRefReleaseVisitor<RAY_TYPE> release_visitor;

            mNegativeEntries.modify( release_visitor );

			mPositiveDegEntries.clear();
            mPositiveNonDegEntries.clear();
            mNegativeEntries.clear();
            mZeroEntries.clear();
		}

        BitSplitVisitor * createBitSplitVisitor ( const TreeOptions & options,
                                                  RayReferenceContainer & container,
                                                  RayReferenceContainer & query_data ) const {

            BitSplitVisitor * visitor;

            if ( options.strategy() == BINARY_STATIC_VANTAGE_POINT ) {
                if ( container.pop_cnt_range() != 0 ) {
                    std::vector<BitList> points( 3, BitList( mBlocks, ULONG_MAX ) );

                    visitor = new BinaryStaticVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), points );
                    visitor->setSuccessor( new BinaryDynamicVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), options.unique_vp() ) );
                } else {
                    visitor = new BinaryDynamicVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), options.unique_vp() );
                }
            } else if ( options.strategy() == BINARY_DYNAMIC_VANTAGE_POINT ) {
                visitor = new BinaryDynamicVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), options.unique_vp() );
            } else {
                BitSelectStrategy * select_strategy;

                switch ( options.strategy() ) {
                    case BPT_MOST_COMMON_BIT_FIRST:
                        select_strategy = new MostCommonBitFirstStrategy( query_data, mPlane );
                        break;
                    case BPT_LEAST_COMMON_BIT_FIRST:
                        select_strategy = new LeastCommonBitFirstStrategy( container, mPlane );
                        break;
                    case BPT_MOST_AVERAGE_BIT_FIRST:
                        select_strategy = new MostAverageBitFirstStrategy( container, mPlane );
                        break;
                    case BPT_LEAST_AVERAGE_BIT_FIRST:
                        select_strategy = new LeastAverageBitFirstStrategy( container, mPlane );
                        break;
                    case BPT_QUERY_BITS_NEUTRALIZATION:
                        select_strategy = new NarrowQueryBitsNeutralizationStrategy( container, query_data, mPlane );
                        break;
                    default:
                        select_strategy = new FirstAvailableBitStrategy( mPlane );
                }

                visitor = new SingleBitSplitVisitor( options.max_tree_depth(),
                                                     options.leaf_capacity(),
                                                     *select_strategy );

                delete select_strategy;
            }

            return visitor;
        }

        BitSplitVisitor * createBitSplitVisitor ( const TreeOptions & options,
                                                  const RayReferenceContainer & container,
                                                  const RayReferenceContainer & pos_container,
                                                  const RayReferenceContainer & neg_container ) const {

            BitSplitVisitor * visitor;

            if ( options.strategy() == BINARY_STATIC_VANTAGE_POINT ) {
                if ( container.pop_cnt_range() != 0 ) {
                    std::vector<BitList> points( 3, BitList( mBlocks, ULONG_MAX ) );

                    visitor = new BinaryStaticVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), points );
                    visitor->setSuccessor( new BinaryDynamicVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), options.unique_vp() ) );
                } else {
                    visitor = new BinaryDynamicVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), options.unique_vp() );
                }
            } else if ( options.strategy() == BINARY_DYNAMIC_VANTAGE_POINT ) {
                visitor = new BinaryDynamicVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), options.unique_vp() );
            } else {
                BitSelectStrategy * select_strategy;

                switch ( options.strategy() ) {
                    case BPT_MOST_COMMON_BIT_FIRST:
                        select_strategy = new MostCommonBitFirstStrategy( container, mPlane );
                        break;
                    case BPT_LEAST_COMMON_BIT_FIRST:
                        select_strategy = new LeastCommonBitFirstStrategy( container, mPlane );
                        break;
                    case BPT_MOST_AVERAGE_BIT_FIRST:
                        select_strategy = new MostAverageBitFirstStrategy( container, mPlane );
                        break;
                    case BPT_LEAST_AVERAGE_BIT_FIRST:
                        select_strategy = new LeastAverageBitFirstStrategy( container, mPlane );
                        break;
                    case BPT_QUERY_BITS_NEUTRALIZATION:
                        select_strategy = new VerQueryBitsNeutralizationStrategy( container, pos_container, neg_container, mPlane );
                        break;
                    default:
                        select_strategy = new FirstAvailableBitStrategy( mPlane );
                }

                visitor = new SingleBitSplitVisitor( options.max_tree_depth(),
                                                     options.leaf_capacity(),
                                                     *select_strategy );

                delete select_strategy;
            }

            return visitor;
        }

        void sortData ( RayReferenceContainer & container,
                        const SortType & type ) {
            
            switch ( type ) {
                case MAX_SET_BITS:
                    container.sort<RefComparatorMostBitsFirst>( RefComparatorMostBitsFirst() );
                    break;
                case MIN_SET_BITS:
                    container.sort<RefComparatorLeastBitsFirst>( RefComparatorLeastBitsFirst() );
                    break;
                case MAX_SET_BITS_LEX:
                    container.sort<RefComparatorMostBitsFirstLex>( RefComparatorMostBitsFirstLex() );
                    break;
                case MIN_SET_BITS_LEX:
                    container.sort<RefComparatorLeastBitsFirstLex>( RefComparatorLeastBitsFirstLex() );
                    break;
                case MAX_BIT_FREQUENCY:
                    calcFrequency( container, 1 );
                    container.sort<RefComparatorMaxFrequency>( RefComparatorMaxFrequency() );
                    break;
                case MIN_BIT_FREQUENCY:
                    calcFrequency( container, 1 );
                    container.sort<RefComparatorMinFrequency>( RefComparatorMinFrequency() );
                    break;
                case MAX_SET_BITS_MAX_BIT_FREQUENCY:
                    calcFrequency( container, 1 );
                    container.sort<RefComparatorMaxSetBitsMaxFrequency>( RefComparatorMaxSetBitsMaxFrequency() );
                    break;   
                case MAX_SET_BITS_MIN_BIT_FREQUENCY:
                    calcFrequency( container, 1 );
                    container.sort<RefComparatorMaxSetBitsMinFrequency>( RefComparatorMaxSetBitsMinFrequency() );
                    break;       
                case MIN_SET_BITS_MIN_BIT_FREQUENCY:
                    calcFrequency( container, 1 );
                    container.sort<RefComparatorMinSetBitsMinFrequency>( RefComparatorMinSetBitsMinFrequency() );
                    break; 
                case MIN_SET_BITS_MAX_BIT_FREQUENCY:
                    calcFrequency( container, 1 );
                    container.sort<RefComparatorMinSetBitsMaxFrequency>( RefComparatorMinSetBitsMaxFrequency() );
                    break; 
                case NONE:
                    /* Do nothing here */
                    break;    
            }
        } 

        void calcFrequency ( RayReferenceContainer & container,
                             const bool value ) {

            BitCountVisitor bit_counter( mPlane );
            container.iterate( bit_counter );
            if ( value ) {
                const BitFrequencyModifier modifier( bit_counter.population());
                container.modify( modifier );
            } else {
                const ZeroFrequencyModifier modifier( bit_counter.population(), container.size() );
                container.modify( modifier );
            }
        }

        void calcFrequency ( RayReferenceContainer & container,
                             RayReferenceContainer & query_data,
                             const bool value ) const {

            BitCountVisitor bit_counter( mPlane );
            query_data.iterate( bit_counter );
            if ( value ) {
                const BitFrequencyModifier modifier( bit_counter.population());
                container.modify( modifier );
            } else {
                const ZeroFrequencyModifier modifier( bit_counter.population(), container.size() );
                container.modify( modifier );
            }
        }

        const TreeNode * createNarrowingTree ( RayReferenceContainer & first,
                                                                     RayReferenceContainer & second,
                                                                     const unsigned short minCommonBits,
                                                                     const TreeOptions & options ) {
         
            TreeNode * tree;

            if ( first.size() > options.leaf_capacity() ) {

                sortData ( first, options.data_sort() );

                BitSplitVisitor * split_visitor = createBitSplitVisitor( options, first, second );
    #ifdef __USE_TEMPLATES__
                if ( split_visitor->supportsExtendedCondition() ) {
                   IntermediateBitSetNode<VantagePointPreliminaryCondition> * root = new IntermediateBitSetNode<VantagePointPreliminaryCondition>( first.union_map(),
                                                                                         VantagePointPreliminaryCondition( first.union_map(), minCommonBits, BitList( mBlocks )/*, Interval(0, 0), Interval( first.min_pop_cnt(), first.max_pop_cnt() )*/ ),
                                                                                         split_visitor->getSplitFactor() );
                   root->partition( first, *split_visitor );
                   tree = root;
                } else {
                   IntermediateBitSetNode<PreliminaryCondition> * root = new IntermediateBitSetNode<PreliminaryCondition>( first.union_map(),
                                                                             PreliminaryCondition( first.union_map(), minCommonBits ),
                                                                             split_visitor->getSplitFactor() );
                   root->partition( first, *split_visitor );   
                   tree = root;                 
                }
    #else // Use polymorphism
			    IntermediateBitSetNode * root = new IntermediateBitSetNode( first.union_map(),
                                                                             new PreliminaryCondition( first.union_map(), minCommonBits ),
                                                                             split_visitor->getSplitFactor() );
                root->partition( first, *split_visitor );
                tree = root;
    #endif

                delete split_visitor;
            } else {
    #ifdef __USE_TEMPLATES__
                tree = new LeafNode<PreliminaryCondition>( first,
                                                           PreliminaryCondition( first.union_map(), minCommonBits ));
    #else // Use polymorphism
                tree = new LeafNode( first,
                                     new PreliminaryCondition( first.union_map(), minCommonBits ));
    #endif
            }

            return tree;
        }


		const struct StepStats fillCandidatesBuffer ( std::vector<const RAY_TYPE*> & buffer,
                                                      const unsigned short minCommonBits ) {

            struct StepStats stats;

            if ( mPositiveDegEntries.size() + mPositiveNonDegEntries.size() > 0 && !mNegativeEntries.empty() ) {

                const Options & options = Options::current();

                sortData ( mZeroEntries, options.verification().data_sort() );

    #ifdef __STATS_OUTPUT__
                PreliminaryCondition::start_new_stat_session();
                VerificationCondition::start_new_stat_session();
    #endif

                std::deque<struct EdgeCandidate> candidates;
                std::deque<struct EdgeCandidate> verified_candidates;
        
                if ( !mPositiveDegEntries.empty() ) {
                    const TreeNode * verificator = createVerTree();
                    const TreeNode * pos_deg_tree = createNarrowingTree( mPositiveDegEntries, mNegativeEntries, minCommonBits, options.deg_narrowing() ); 

                    VerificationTraverseVisitor verification_visitor( *pos_deg_tree, *verificator, minCommonBits );
                    mNegativeEntries.iterate( verification_visitor );
                    candidates = verification_visitor.candidates();

                    delete verificator;
                    delete pos_deg_tree;
                }

                if ( !mPositiveNonDegEntries.empty() ) {
                    const TreeNode * pos_nondeg_tree = createNarrowingTree( mPositiveNonDegEntries, mNegativeEntries, minCommonBits, options.non_deg_narrowing() );

                    PreliminaryTraverseVisitor narrowing_visitor( *pos_nondeg_tree, minCommonBits );
                    mNegativeEntries.iterate( narrowing_visitor );
                    verified_candidates = narrowing_visitor.candidates();

                    delete pos_nondeg_tree;
                }

    #ifdef __STATS_OUTPUT__
                stats.elimOps = PreliminaryCondition::NARROWING_OPS;
                double n_log_n_complexity = mNegativeEntries.size() * log2( mPositiveDegEntries.size() + mPositiveNonDegEntries.size() );
                double quad_complexity = mNegativeEntries.size() * ( mPositiveDegEntries.size() + mPositiveNonDegEntries.size() );
                stats.elimComplexity = ( stats.elimOps - n_log_n_complexity ) / ( quad_complexity - n_log_n_complexity ) + 1; 

                stats.verificationCandidates = VerificationCondition::CANDIDATES;    
                stats.verOps = VerificationCondition::OPS + PreliminaryCondition::VERIFICATION_OPS;
                stats.verified = candidates.size();
            
                if ( !mZeroEntries.empty() && stats.verificationCandidates > 0 ) {
                    double log_n_complexity = log2( mZeroEntries.size() ) * stats.verificationCandidates;
                    double log_n_3_complexity = log2( mZeroEntries.size() ) * log2( mZeroEntries.size() ) * log2( mZeroEntries.size()  ) * stats.verificationCandidates;
                    double n_complexity = mZeroEntries.size() * stats.verificationCandidates;
                    stats.verComplexityLg = ( stats.verOps - log_n_complexity )/ ( log_n_3_complexity - log_n_complexity ) + 1; 
                    stats.verComplexityLin = ( stats.verOps - log_n_complexity )/ ( n_complexity - log_n_complexity ) + 1;        
                }

                DegeneracyCalcVisitor degeneracy_visitor( minCommonBits + 1 );

                if ( !mPositiveDegEntries.empty() || !mPositiveNonDegEntries.empty() ) {
                    mPositiveDegEntries.iterate( degeneracy_visitor );
                    stats.degeneracyPos = degeneracy_visitor.degeneracy() * mPositiveDegEntries.size() /( mPositiveDegEntries.size() + mPositiveNonDegEntries.size() );
                } else {
                    stats.degeneracyPos = ULONG_MAX;
                }

                if ( !mNegativeEntries.empty() ) {
                    mNegativeEntries.iterate( degeneracy_visitor );
                    stats.degeneracyNeg = degeneracy_visitor.degeneracy();   
                } else {
                    stats.degeneracyNeg = ULONG_MAX;
                }
    #endif

                transform( verified_candidates, buffer );
                transform( candidates, buffer );
            }

			return stats;
		}

        const TreeNode * createVerTree () const {

                const TreeNode * verificator;
                const Options & options = Options::current();

                if ( !mZeroEntries.empty() ) {
                    BitSplitVisitor * zero_visitor = createBitSplitVisitor( options.verification(),
                                                                            mZeroEntries,
                                                                            mPositiveDegEntries,
                                                                            mNegativeEntries );
    #ifdef __USE_TEMPLATES__
                    if ( zero_visitor->supportsExtendedCondition() ) {
                        const VantagePointPreliminaryCondition empty_cond( mZeroEntries.union_map(), mMinCommonBits, BitList( mBlocks ) );
                        IntermediateBitSetNode<VantagePointPreliminaryCondition> * zero_tree = new IntermediateBitSetNode<VantagePointPreliminaryCondition>( mZeroEntries.union_map(),
                                                                                                                                                             empty_cond,
                                                                                                                                                             zero_visitor->getSplitFactor() );
                        zero_tree->partition( mZeroEntries, *zero_visitor );
                        verificator = zero_tree;   
                    } else {
                        IntermediateBitSetNode<PreliminaryCondition> * zero_tree = new IntermediateBitSetNode<PreliminaryCondition>( mZeroEntries.union_map(),
                                                                                                                                     PreliminaryCondition( mZeroEntries.union_map(), mMinCommonBits ),
                                                                                                                                     zero_visitor->getSplitFactor() ); 
                        zero_tree->partition( mZeroEntries, *zero_visitor );  
                        verificator = zero_tree;        
                    }
    #else // Use polymorphism
                    IntermediateBitSetNode * zero_tree = IntermediateBitSetNode( mZeroEntries.union_map(), new PreliminaryCondition( mZeroEntries.union_map(), mMinCommonBits ), 2 );
                    zero_tree->partition( mZeroEntries, *zero_visitor );
                    verificator = zero_tree;      
    #endif
                    
                    delete zero_visitor;

                } else {
    #ifdef __USE_TEMPLATES__
                    verificator = & LeafNode<PreliminaryCondition>::EMPTY_NODE;
    #else
                    verificator = LeafNode::EMPTY_NODE;
    #endif
                }

                return verificator;
        }

#ifdef __COMBINATIONAL_TEST__

        void transform ( const std::deque<struct EdgeCandidate> & verified,
                         std::vector<const RAY_TYPE*> & buffer ) const {

            for ( unsigned long index = 0; index < verified.size(); index++ ) {
			    const RAY_TYPE * first = static_cast<const RAY_TYPE*>( verified[index].first.ray );
			    const RAY_TYPE * second = static_cast<const RAY_TYPE*>( verified[index].second.ray );

     		    const RAY_TYPE * positive = first->isInside( mPlane ) ? first : second;
				const RAY_TYPE * negative = first->isInside( mPlane ) ? second : first;                   

			    buffer.push_back( new SingleSetRay( positive, negative, mPlane, mRows ));
            }
        }

        void findAdjacentRays ( const std::deque<struct EdgeCandidate> & candidates,
                                const TreeNode & current,
                                std::vector<const RAY_TYPE*> & buffer ) const {

			for ( unsigned long index = 0; index < candidates.size(); index++ ) {
 
				const ExtVerificationCondition ver_cond( candidates[ index ].activeSet,  candidates[ index ].first.activeSet,  candidates[ index ].second.activeSet );
                
                if ( !current.verificationTraverse( ver_cond ) ) {

				    const RAY_TYPE * first = static_cast<const RAY_TYPE*>( candidates[index].first.ray );
				    const RAY_TYPE * second = static_cast<const RAY_TYPE*>( candidates[index].second.ray );

     				const RAY_TYPE * positive = first->isInside( mPlane ) ? first : second;
				    const RAY_TYPE * negative = first->isInside( mPlane ) ? second : first;                   

					buffer.push_back( new SingleSetRay( positive, negative, mPlane, mRows ));                    
                }
            }
        }

#else // __ALGEBRAIC_TEST__

        void findAdjacentRays ( const std::deque<struct EdgeCandidate> & candidates,
                                std::vector<const RAY_TYPE*> & buffer ) const {

            for ( unsigned long index = 0; index < candidates.size(); index++ ) {

				const RAY_TYPE * first = static_cast<const RAY_TYPE*>( candidates[index].first.ray );
				const RAY_TYPE * second = static_cast<const RAY_TYPE*>( candidates[index].second.ray );

     		    const RAY_TYPE * positive = first->isInside( mPlane ) ? first : second;
				const RAY_TYPE * negative = first->isInside( mPlane ) ? second : first; 

				buffer.push_back( positive );
				buffer.push_back( negative );
            }
        }
#endif // __COMBINATIONAL_TEST__

	};
}

#endif // __DD_BIT_PATTERN_TREE_CACHE_ENTRY__
