/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_BIT_SPLIT_VISITOR__
#define __DD_BIT_SPLIT_VISITOR__

#include <vector>

#include "DDBitList.h"
#include "DDRayRefVisitor.h"

#include "DDRayReferenceContainer.h"

#ifdef __USE_TEMPLATES__
    #include "DDPreliminaryCondition.h"
    #include "DDVantagePointPreliminaryCondition.h"
#endif

namespace double_description {

    enum BitSplitVisitorState {
        TERMINATE,
        CHANGE_VISITOR,
        CONTINUE
    };

    class BitSplitVisitor : public RayRefVisitor {

        protected:
        unsigned short mCurrentDepth;

        const unsigned short mMaxDepth;
        const unsigned short mLeafCapacity;

        RayReferenceContainer * mSets;
        unsigned short mSetsCount;

        BitSplitVisitor * mSuccessor;
  
        long mCurrentEntry;
        unsigned long mEntriesCount;

        RayReferenceContainer & set ( unsigned short index ) {
            return mSets[ index ];
        }

        const RayReferenceContainer & set ( unsigned short index ) const {
            return mSets[ index ];
        }

        unsigned short current_depth () const {
            return mCurrentDepth;
        }

        public:
        BitSplitVisitor( const unsigned short depth,
                         const unsigned short capacity,
                         const unsigned short split ) :
            RayRefVisitor(),
            mCurrentDepth( 0 ),
            mMaxDepth( depth ),
            mLeafCapacity( capacity ),
            mSetsCount( split ),
            mSuccessor( NULL ) {}

        virtual ~BitSplitVisitor () {
            if ( hasSuccessor() ) {
                delete mSuccessor;
            }
        }

        virtual void onStart () {
            RayRefVisitor::onStart();

            for ( unsigned short i = 0; i < mSetsCount; i++ ) {
                set( i ).clear();
            }

            mCurrentEntry = -1;
        }

        virtual void onVisit( const RayReference & ref ) {
            RayRefVisitor::onVisit( ref );
            mCurrentEntry++;
        }

        void registerOutputSets ( RayReferenceContainer * sets ) {
            mSets = sets;
        }

        virtual void before ( const RayReferenceContainer & container ) {
            mEntriesCount = container.size();
            mCurrentDepth++;
        }

        virtual void after () {
            mCurrentDepth--;
        }      

        virtual bool can_proceed ( const unsigned short ) const {
            return true; // mCurrentDepth < mMaxDepth && set( index ).size() > mLeafCapacity;
        }

        virtual bool can_proceed ( const RayReferenceContainer & ) const {
            return true; // mCurrentDepth < mMaxDepth && set( index ).size() > mLeafCapacity;
        }

        virtual const PreliminaryCondition * createNarrowingCondition ( const unsigned short index,
                                                                        const unsigned short bits ) const {
            return new PreliminaryCondition( set( index ).union_map(), bits );
        }

    #ifdef __USE_TEMPLATES__     
        void refineCondition ( const unsigned int index,
                               const unsigned short bits,
                               PreliminaryCondition & cond ) const {
            cond.set( set( index ).union_map() );
            cond.set( bits );
        }

        virtual void refineCondition ( const unsigned int index,
                                       const unsigned short bits,
                                       VantagePointPreliminaryCondition & cond ) const {
            cond.set( set( index ).union_map() );
            cond.set( bits );
        }

        virtual bool supportsExtendedCondition () {
            return false;    
        }
    #endif

        unsigned short getSplitFactor () const {
            return mSetsCount;
        }

        void setSuccessor ( BitSplitVisitor * successor ) {
            mSuccessor = successor;
        }

        BitSplitVisitor & successor () {
            return *mSuccessor;
        }

        const BitSplitVisitor & successor () const {
            return *mSuccessor;
        }

        bool hasSuccessorToProceed ( const RayReferenceContainer & container ) const {
            return hasSuccessor() && ( successor().can_proceed( container ) || successor().hasSuccessorToProceed( container ));
        }

        BitSplitVisitor & successor_to_proceed ( const RayReferenceContainer & container ) {
            if ( successor().can_proceed( container ) ) {
                return successor();
            } else {
                return successor().successor_to_proceed( container );
            }
        }

        bool hasSuccessor() const {
            return mSuccessor != NULL ;
        }

        bool leaf_condition ( const unsigned short index ) const {
            return mCurrentDepth >= mMaxDepth || set( index ).size() <= mLeafCapacity;
        }

        BitSplitVisitorState state ( const unsigned short index ) const {
            const bool proceed = can_proceed( index );

            if ( leaf_condition( index ) || unlikely( !proceed && !hasSuccessorToProceed( set( index ) ) ) ) {
                return TERMINATE;
            } else if ( !proceed ) {
                return CHANGE_VISITOR;
            } else {
                return CONTINUE;
            }
        }
    };
}

#endif // __DD_BIT_SPLIT_VISITOR__
