/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_CALCULATION_MATRIX__
#define __DD_CALCULATION_MATRIX__

#include "DDBitList.h"
#include "DDMatrix.h"
#include "DDUtilsFactory.h"

#include "DDLogging.h"

namespace double_description {

	class CalculationMatrix : protected Matrix {
		
		private:
		unsigned long mRowLimit;
		unsigned long mColumnLimit;

		protected:
		CalculationMatrix ( const unsigned long rows, const unsigned long columns );
		virtual ~CalculationMatrix () {}

		void transformToUpperTriangularMatrix () {
			unsigned long row, column;

			for ( row = 0, column = 0; ( row < getRows() - 1 ) && ( column < getColumns() - 1 ); column++ ) {
				if ( findPivot( row, column ) ) {
					prepareRectangleForElimination( row, column );	
					eliminateColumnRectangle( row, getRows(), column );
					row++;
				}
			}
		}

		void transformToDiagonalMatrix () {

			for ( long row = getRows() - 2; row >= 0; row-- ) {

				if ( row > 0 ) {
					prepareTriangleForElimination( 0, row + 1, row );
					eliminateColumnTriangle( 0, row + 1 );
				}

				if ( getRow( row ).isNonZeroAt( row ) ) {

					UtilsFactory::CALC_UTILS.startIncrementalGCD( getValue( row, row ) );

					for ( unsigned long column = getRows() - 1; column < getColumns(); column++ ) {
						if ( getRow( row ).isNonZeroAt( column ) ) {
							UtilsFactory::CALC_UTILS.calcIncrementalGCD( getValue( row, column ) );
						}
					}

					getRow( row ).divide( row, UtilsFactory::CALC_UTILS.getResult() );
					getRow( row ).divide( getRows() - 1, getColumns(), UtilsFactory::CALC_UTILS.getResult() );			
				} 
			}
		}

		void prepareRectangleForElimination ( const unsigned long beginRow, const unsigned long column ) {
			prepareRectangleForElimination( beginRow, getRows(), column );
		}

		void prepareRectangleForElimination ( const unsigned long beginRow, const unsigned long endRow, const unsigned long column ) {

			UtilsFactory::CALC_UTILS.startIncrementalLCM ( getValue( beginRow, column ) );

			for ( unsigned long i = beginRow + 1; i < endRow; i++ ) {
				UtilsFactory::CALC_UTILS.calcIncrementalLCM ( getValue ( i, column ) );
			}
			UtilsFactory::CALC_UTILS.convertResultToNegativeValue();
			multiplyRectangle( beginRow, endRow, column, UtilsFactory::CALC_UTILS.getResult() );
		}

		void prepareTriangleForElimination ( const unsigned long beginRow, const unsigned long endRow, const unsigned long pivotColumn ) {

			UtilsFactory::CALC_UTILS.startIncrementalLCM ( getValue( beginRow, pivotColumn ) );

			for ( unsigned long i = beginRow + 1; i < endRow; i++ ) {
				UtilsFactory::CALC_UTILS.calcIncrementalLCM ( getValue ( i, pivotColumn ) );
			}
			UtilsFactory::CALC_UTILS.convertResultToNegativeValue();
			multiplyUpperTriangle( beginRow, endRow, pivotColumn, UtilsFactory::CALC_UTILS.getResult() );	
		}

		void eliminateColumnTriangle ( const unsigned long beginRow, const unsigned long endRow ) {

			for ( unsigned long index = beginRow; index < endRow - 1; index++ ) {
				if ( getValue( index, endRow - 1 ) != 0 ) {
					getRow( index ).add( endRow - 1, getRow( endRow - 1 ) );
					getRow( index ).add( getRows() - 1, getColumns(), getRow( endRow - 1 ) );
				}
			}
		}

		void eliminateColumnRectangle ( const unsigned long beginRow, const unsigned long endRow, const unsigned long beginColumn ) {

			for ( unsigned long index = beginRow + 1; index < endRow; index++ ) {
				if ( getValue( index, beginColumn ) != 0 ) {
					getRow( index ).add( beginColumn, getColumns(), getRow( beginRow ) );
				}
			}
		}

		void multiplyRectangle ( const unsigned long beginRow, const unsigned long pivotColumn, const mpz_class& lcm ) {
			multiplyRectangle( beginRow, getRows(), pivotColumn, getColumns(), lcm );
		}

		void multiplyRectangle ( const unsigned long beginRow, const unsigned long endRow, const unsigned long pivotColumn, const mpz_class& lcm ) {
			multiplyRectangle( beginRow, endRow, pivotColumn, getColumns(), lcm );
		}

		void multiplyRectangle ( const unsigned long beginRow, const unsigned long endRow, const unsigned long pivotColumn, const unsigned long endColumn, const mpz_class& lcm ) {

			UtilsFactory::CALC_UTILS.loadRegisterS( lcm );
			UtilsFactory::CALC_UTILS.divideRegisterS( getValue( beginRow, pivotColumn ) );
			UtilsFactory::CALC_UTILS.negateRegisterS();
			getRow( beginRow ).multiply( pivotColumn, endColumn, UtilsFactory::CALC_UTILS.getRegisterS() );

			for ( unsigned long row = beginRow + 1; row < endRow; row++ ) {
				if ( getRow( row ).isNonZeroAt( pivotColumn ) && lcm != getValue( row, pivotColumn ) ) {
					UtilsFactory::CALC_UTILS.loadRegisterS( lcm );
					UtilsFactory::CALC_UTILS.divideRegisterS( getValue( row, pivotColumn ) );		
					getRow( row ).multiply( pivotColumn, endColumn, UtilsFactory::CALC_UTILS.getRegisterS() );
				}
			}
		}

		void multiplyUpperTriangle ( const unsigned long beginRow, const unsigned long endRow, const unsigned long pivotColumn, const mpz_class& lcm ) {
			
			for ( unsigned long row = beginRow; row < endRow - 1; row++ ) {
				if ( getRow( row ).isNonZeroAt( pivotColumn ) && lcm != getValue( row, pivotColumn ) ) {
					UtilsFactory::CALC_UTILS.loadRegisterS( lcm );
					UtilsFactory::CALC_UTILS.divideRegisterS( getValue( row, pivotColumn ) );
					getRow( row ).multiply( row, getColumns(), UtilsFactory::CALC_UTILS.getRegisterS() );
				}	
			}	

			UtilsFactory::CALC_UTILS.loadRegisterS( lcm );
			UtilsFactory::CALC_UTILS.divideRegisterS( getValue( endRow - 1, pivotColumn ) );
			UtilsFactory::CALC_UTILS.negateRegisterS();
			getRow( endRow - 1 ).multiply( endRow - 1, getColumns(), UtilsFactory::CALC_UTILS.getRegisterS() );
		}

		bool isSolvable () const {
			bool result = true;
				
			for ( unsigned long row = getColumns() - 1; row < getRows(); row++ ) {
				if ( isNonZeroAt( row, getColumns() - 1 ) ) {
					result = false;
					break;
				} 
			}

			return result; 
		}

		bool findPivot ( const unsigned long beginRow, const unsigned pivotColumn ) {
			bool result = false;

			if ( !isNonZeroAt( beginRow, pivotColumn ) ) {
				for ( unsigned long row = beginRow + 1; row < getRows(); row++ ) {
					if ( isNonZeroAt( row, pivotColumn ) ) {
						swapRows( row, beginRow );
						result = true;
						break;
					}
				}		
			} else {
				result = true;
			}

			return result;
		}

		void setRows ( const unsigned long rows ) {
			mRowLimit = rows;
		}

		void setColumns ( const unsigned long columns ) {
			mColumnLimit = columns;
		}

		public:

		unsigned long getRows () const {
			return mRowLimit;
		}

		unsigned long getColumns () const {
			return mColumnLimit;
		}

		void simplify () {
            unsigned long first_row;

            for ( first_row = 0; !isNonZeroAt( first_row, first_row ) && first_row < getRows() - 1; first_row++ );

			UtilsFactory::CALC_UTILS.startIncrementalLCM ( getValue( first_row, first_row ) );

			for ( unsigned long row = first_row + 1; row < getRows() - 1; row++ ) {
					UtilsFactory::CALC_UTILS.calcIncrementalLCM ( getValue( row, row ) );
			}

			for ( unsigned long row = first_row; row < getRows() - 1; row++ ) {
					UtilsFactory::CALC_UTILS.loadRegisterS( UtilsFactory::CALC_UTILS.getResult() );
					UtilsFactory::CALC_UTILS.divideRegisterS( getValue( row, row ) );

					getRow( row ).multiply( row, UtilsFactory::CALC_UTILS.getRegisterS() );
					getRow( row ).multiply( getRows() - 1, getColumns(), UtilsFactory::CALC_UTILS.getRegisterS() );
			} 
		}

		const std::string toString () const;
	};
}

#endif // __DD_LINEAR_DEPENDENCE_CALCULATOR__
