/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_LEAF_NODE__
#define __DD_LEAF_NODE__

#include "DDTreeNode.h"
#include "DDPreliminaryCondition.h"
#include "DDVantagePointPreliminaryCondition.h"
#include "DDVerificationCondition.h"

#include "DDRayReferenceContainer.h"

#include "DDRayRefContainerToArrayTransformer.h"

namespace double_description {

    #ifdef __USE_TEMPLATES__
    template<class CONDITION_TYPE>
    #endif
	class LeafNode : virtual public TreeNode {

        private:
        LeafNode ();

		protected:
    #ifdef __USE_TEMPLATES__    
        const CONDITION_TYPE mCondition;
    #endif
		const unsigned long mRefsCount;
		const RayReference * mRefs;

		static const RayReference * createReferences ( const RayReferenceContainer & refs ) {
			RayRefContainerToArrayTransformer transformer( refs.size() );

			refs.iterate( transformer );		

			return transformer.data();	
		}

		bool checkRays ( const ExtVerificationCondition & condition ) const {

			for ( unsigned int index = 0; index < mRefsCount; index++ ) {
				if ( condition.holdsOnRay( mRefs[index].activeSet) ) {
					return true;
				}
			}	

			return false;		
		}

		public:
        
    #ifdef __USE_TEMPLATES__ 
        static LeafNode<PreliminaryCondition> EMPTY_NODE;
    #else
        static LeafNode EMPTY_NODE;
    #endif

		LeafNode ( const RayReferenceContainer & candidates,
    #ifdef __USE_TEMPLATES__ 
                   const CONDITION_TYPE & cond
    #else // Use polymorphism
                   const PreliminaryCondition * const cond
    #endif
            ) :
    #ifdef __USE_TEMPLATES__ 
            TreeNode( candidates.union_map() ),
            mCondition( cond ),
    #else // Use polymorphism
	        TreeNode( candidates.union_map(), cond ),
    #endif
	        mRefsCount( candidates.size() ),
	        mRefs( createReferences( candidates ) ) {}

        ~LeafNode () {
            if ( __builtin_expect( mRefs != NULL, true ) ) {
	            delete[] mRefs;
            }
        }

		void preliminaryTraverse ( const TreeNode * partner,
									std::vector<struct EdgeCandidate> & candidates ) const {
            if ( partner->getCondition()->holdsOnFace( mMap ) ) {
	            partner->preliminaryTraverse( mRefs, mRefsCount, *mCond, candidates );
            }
        }

		void preliminaryTraverse (  const RayReference & current,
									std::deque<struct EdgeCandidate> & candidates ) const {

            if (
    #ifdef __USE_TEMPLATES__         
             mCondition.proceed_narrowing( current.activeSet )
    #else // Use polymorphism
             mCond->proceed_narrowing( current.activeSet )
    #endif
            ) {            
	             for ( unsigned int index = 0; index < mRefsCount; index++ ) {
		            if ( 
    #ifdef __USE_TEMPLATES__
                      mCondition.feasable( current.activeSet, mRefs[ index ].activeSet )
    #else // Use polymorphism
                      mCond->feasable( current.activeSet, mRefs[ index ].activeSet )
    #endif 
                    ) {
			            candidates.push_back( EdgeCandidate ( current, mRefs[ index ] ));
		            }
	             }
            }
        }

		void preliminaryTraverse (  const RayReference & current,
									std::deque<struct EdgeCandidate> & candidates,
                                    const TreeNode & verificator ) const {

            if (
    #ifdef __USE_TEMPLATES__         
             mCondition.proceed_narrowing( current.activeSet )
    #else // Use polymorphism
             mCond->proceed_narrowing( current.activeSet )
    #endif
            ) {            
	             for ( unsigned int index = 0; index < mRefsCount; index++ ) {
		            if ( 
    #ifdef __USE_TEMPLATES__
                      mCondition.feasable( current.activeSet, mRefs[ index ].activeSet )
    #else // Use polymorphism
                      mCond->feasable( current.activeSet, mRefs[ index ].activeSet )
    #endif 
                    ) {
                        const ExtVerificationCondition ver_cond( current.activeSet & mRefs[ index ].activeSet, current.activeSet,  mRefs[ index ].activeSet );

                        if ( !verificator.verificationTraverse( ver_cond ) ) {
    			            candidates.push_back( EdgeCandidate ( current, mRefs[ index ] ));
                        }
		            }
	             }
            }
        }

        void preliminaryTraverse ( const RayReference * refs,
									const unsigned int refsCount,
									std::vector<struct EdgeCandidate> & candidates ) const {

	        for ( unsigned int first_i = 0; first_i < mRefsCount; first_i++ ) {
		        PreliminaryCondition ray_cond( mRefs[first_i].activeSet, mCond->getMinCommonBits() );			

		        if ( ray_cond.holdsOnFace( mMap ) ) {
			        for ( unsigned int second_i = 0; second_i < refsCount; second_i++ ) {
				        if ( ray_cond.holdsOnRay( refs[second_i].activeSet ) ) {
					        candidates.push_back( EdgeCandidate ( mRefs[first_i], refs[second_i] ) );
				        }
			        }
		        }	
	        }
        }

		void preliminaryTraverse ( const RayReference * refs,
									const unsigned int refsCount,
									const PreliminaryCondition & condition,
									std::vector<struct EdgeCandidate> & candidates ) const {

	        if ( condition.holdsOnFace( mMap ) ) {
		        preliminaryTraverse( refs, refsCount, candidates );	
	        }
        }

		bool verificationTraverse ( const ExtVerificationCondition & condition ) const {

    #ifdef __USE_TEMPLATES__
            const bool redundant = mCondition.isRedundant( condition );
    #else // Use polymorphism
            const bool redundant = mCond->isRedundant( condition );
    #endif	

            if ( redundant ) {
                return true;
            }

	        if ( 
    #ifdef __USE_TEMPLATES__ 
             mCondition.proceed_verification( condition.getMap() ) 
    #else // Use polymorphism
             mCond->proceed_verification( condition.getMap() ) 
    #endif
            ) {
		        return checkRays( condition );
	        } else {
		        return false;
	        }
        }

        void enumerateLeafNodes( std::vector<const TreeNode*> & container ) const {
            if ( !isEmpty() ) {
                container.push_back( this );
            }
        }

        unsigned long getSubNodesCount () const {
            return 0;
        }

        unsigned long getRaysCount () const {
            return mRefsCount;
        }

        unsigned long getLeafNodesCount () const {
            return isEmpty() ? 0 : 1;
        }

        bool isEmpty () const {
            return getRaysCount () == 0;
        }

        const std::string toString ( const unsigned short ) const {
            std::stringstream result;

            const TreeNode * node = this;
            
            result << "p" << static_cast<const void*>( node )
                   << " [label=\"" << "|V|=" << getRaysCount() << ", "
    #ifdef __USE_TEMPLATES__
                   << mCondition.toString()          
    #else
                   << mCond->toString()
    #endif     
                   << "\"];\n";

            return result.str();
        }
	};

    #ifdef __USE_TEMPLATES__
            template<>
            LeafNode<PreliminaryCondition>::LeafNode();   

            template<class CONDITION_TYPE>
            LeafNode<PreliminaryCondition> LeafNode<CONDITION_TYPE>::EMPTY_NODE;
    #endif
}

#endif // __DD_LEAF_NODE__
