/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "DDMatrix.h"
#include "DDUtilsFactory.h"

#include "DDLogging.h"

#include <sstream>

using namespace double_description;

Matrix::Matrix () {}

Matrix::Matrix ( const unsigned long rows, const unsigned long columns ) {

	for ( unsigned long row_index = 0; row_index < rows; row_index++ ) {
		const CalculationVector& row = CalculationVector( columns );
		mData.push_back( row );
		mMapping.push_back( row_index );
	}	
}

Matrix::Matrix ( const unsigned long rows, const unsigned long columns, const std::vector<unsigned long>& mapping ) :
	mMapping( mapping ) {

	for ( unsigned long row_index = 0; row_index < rows; row_index++ ) {
		const CalculationVector& row = CalculationVector( columns );
		mData.push_back( row );
	}		
}

Matrix::Matrix ( const Matrix & other ) :
	mData( other.mData ),
	mMapping( other.mMapping ) {}

void Matrix::insert ( const CalculationVector& row ) {
	mMapping.push_back( mData.size() );
	mData.push_back( row );
}

Matrix Matrix::operator * ( const Matrix & rhs ) const {
	Matrix result = *this;
	
	result *= rhs;

	return result;
}

void Matrix::operator *= ( const Matrix& matrix ) {

	CalculationVector buffer( matrix.getColumns() );

	for ( unsigned long row = 0; row < getRows(); row++ ) {
		
		for ( unsigned long column = 0; column < matrix.getColumns(); column++ ) {

			buffer.setValue( column, 0 );

			for ( unsigned long index = 0; index < getColumns(); index++ ) {
				if ( isNonZeroAt( row, index ) && matrix.isNonZeroAt( index, column ) ) {

					UtilsFactory::CALC_UTILS.loadRegisterT( getValue( row, index ) );
					UtilsFactory::CALC_UTILS.multiplyRegisterT( matrix.getValue( index, column ) );
					buffer.add( column, UtilsFactory::CALC_UTILS.getRegisterT() );

				}
			}
		}	
		getRow( row ) = buffer; 
	}
}

void Matrix::operator = ( const Matrix& matrix ) {
	mData = matrix.mData;
	mMapping = matrix.mMapping;
}

Matrix Matrix::getSubMatrix ( const unsigned long beginRow, const unsigned long endRow ) const {
	Matrix result;	

	for ( unsigned long index = beginRow; index < endRow; index++ ) {
		result.insert( getRow( index ) );
	}

	return result;
}

void Matrix::optimizeRowsOrder () {

	unsigned long min;

	for ( unsigned long i = getColumns(); i < getRows(); i++ ) {

		min = i;
		for ( unsigned long l = i + 1; l < getRows(); l++ ) {
			if ( getRow( min ).getNonZeroCount() >= getRow( l ).getNonZeroCount() ) {
				min = l;
			}
		}
		swapRows( min, i );
	} 
}


void Matrix::sort ( const SortCriteria & criteria, unsigned short start, unsigned short end, int swap_trigger ) {

	for ( unsigned short i = start; i < end; i++ ) {

		unsigned short pivot = i;
		for ( unsigned long l = i + 1; l < end; l++ ) {
			if ( criteria.cmp ( getRow( l ), getRow( pivot )) == swap_trigger ) {
				pivot = l;
			}
		}
		swapRows( pivot, i );
	} 	
}

const std::string Matrix::getMapping () const {
	std::stringstream output;

	output << "[";
	for ( unsigned long i = 0; i < getRows() - 1; i++ ) {
		output << mMapping[i] << ", ";
	}
	output << mMapping[ getRows()-1 ] << "]";

	return output.str();
}

const std::string Matrix::toString () const {
	std::stringstream output;

	for ( unsigned long i = 0; i < getRows() - 1; i++ ) {
		output << getRow( i ).toString() << std::endl;
	}
	output << getRow( getRows() - 1 ).toString();

	return output.str();
}


