/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_RAY_CACHE__
#define __DD_RAY_CACHE__

#include <deque>

#include "DDMatrix.h"

namespace double_description {

	template <class RAY_TYPE, typename CACHE_ENTRY_TYPE>
	class RayCache {

		private:
		const Matrix & mSystem;
		const Matrix & mTransformator;

        bool mDynamicPolicy;

		std::deque<CACHE_ENTRY_TYPE> mEntries;

#ifdef __COMPRESSION__
		std::vector<const RAY_TYPE*> mCompressedResults;	
#endif // __COMPRESSION__

		std::vector<RepresentationVector> mResults;	

		short insert_local ( const RAY_TYPE * const ray,
                             const unsigned long index,
                             const bool insert_zero ) {

            const unsigned long plane = mSystem.getRows() - entries() + index;
        
		    if ( ray->isInside( plane) ) {
			    mEntries[index].insertPositive( ray );
                return 1;
			} else if ( ray->isOutside( plane ) ) {
			    mEntries[index].insertNegative( ray );
                return -1;
	        } else {
                if ( insert_zero ) {
			        mEntries[index].insertZero( ray );
                }
                return 0;
			}
        }

		public:
		RayCache ( const Matrix & system, const Matrix & transformator ) :
			mSystem ( system ),
			mTransformator( transformator ),
            mDynamicPolicy( false ) {

			for ( unsigned short p = system.getColumns(); p < system.getRows(); p++ )
#ifdef __BIT_PATTERN_TREES__
				mEntries.push_back( CACHE_ENTRY_TYPE( system.getRows(), system.getColumns(), p ) );
#else
	#ifdef __POPULATION_TREES__
				mEntries.push_back( CACHE_ENTRY_TYPE( p, system.getColumns() - 2 )); 
	#else
				mEntries.push_back( CACHE_ENTRY_TYPE( p, system.getColumns()/2 - 1 ));
	#endif // __POPULATION_TREES__
#endif // #ifdef __BIT_PATTERN_TREES__
		}

		~RayCache () {
#ifdef __COMPRESSION__
            for ( unsigned long i = 0; i < mCompressedResults.size(); i++ ) {
		        delete mCompressedResults[i];	
            }
#endif // __COMPRESSION__
        }

        void insert ( const RAY_TYPE * const ray ) {
            insert( ray, 0 );
        }

		void insert ( const RAY_TYPE * const ray, const unsigned long step ) {
	
			bool final = true;

            if ( mDynamicPolicy ) {
                if ( step < mEntries.size() ) {
                    insert_local( ray, step, true );
                    final = false;
                }
            } else {
#ifdef __COMBINATIONAL_TEST__
                bool insert_zero = true;
#else
                bool insert_zero = false;
#endif // __COMBINATIONAL_TEST__
			    for ( unsigned long index = step; index < mEntries.size(); index++ ) {

				    if ( insert_local( ray, index, insert_zero ) < 0 ) {
					    final = false;
					    break;
                    }
				}
			}

			if ( final ) {
#ifdef __COMPRESSION__
				mCompressedResults.push_back( ray );
#else
                RepresentationVector vec = *ray;
                vec.simplify();
				mResults.push_back( vec );
#endif // __COMPRESSION__
			}
		}

		const CACHE_ENTRY_TYPE & getEntry ( const unsigned long index ) const {
			return mEntries[index];
		}

		CACHE_ENTRY_TYPE & getEntry ( const unsigned long index ) {
			return mEntries[index];
		}       

        void pop () {
            mEntries.pop_front();
        }       

        const CACHE_ENTRY_TYPE & next () const {
            return mEntries.front();
        }

        CACHE_ENTRY_TYPE & next () {
            return mEntries.front();
        }

        unsigned short entries () const {
            return mEntries.size();
        }

#ifdef __COMPRESSION__
		void uncompress () {
			mResults.resize( mCompressedResults.size(), RepresentationVector( mSystem.getColumns() ) );
		}

		const std::vector<const RAY_TYPE*> & getCompressedResults () const {
			return mCompressedResults;
		}

		std::vector<const RAY_TYPE*> & getCompressedResults () {
			return mCompressedResults;
		}
#endif

		unsigned long getExtremeRays () const {
#ifdef __COMPRESSION__
			return mCompressedResults.size();
#else
			return mResults.size();
#endif
		}

		const std::vector<RepresentationVector>& getResults () const {
			return mResults;
		}
	};
}

#endif // __DD_RAY_CACHE__
