/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_RAY_REFERENCE__
#define __DD_RAY_REFERENCE__

#include <math.h>

#include "DDBitList.h"

namespace double_description {

	struct RayReference {
		const void * ray;
		BitList activeSet;
        unsigned short degeneracy;
        unsigned long frequency;

		RayReference() :
			ray( NULL ),
			activeSet(),
            degeneracy( 0 ),
            frequency( 0 ) {}

		RayReference( const void * r, const BitList & as ) :
			ray( r ),
			activeSet( as ),
            degeneracy( as.getPopulationCount() ),
            frequency( 0 ) {}

		void operator= ( const RayReference & other ) {
			ray = other.ray;
			activeSet = other.activeSet;
            degeneracy = other.degeneracy;
            frequency = other.frequency;
		}
	};

    struct EdgeCandidate {
        struct RayReference first;
        struct RayReference second;
        BitList activeSet;

        EdgeCandidate( const struct RayReference & ref_1, const struct RayReference & ref_2 ) :
            first( ref_1),
            second( ref_2 ),
            activeSet( ref_1.activeSet & ref_2.activeSet ) {}

        void operator= ( const EdgeCandidate & other ) {
            first = other.first;
            second = other.second;
            activeSet = other.activeSet;
        }
    };

    struct RefComparatorMostBitsFirst {

        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy > rhs.degeneracy;
        }
    };

    struct RefComparatorLeastBitsFirst {

        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy < rhs.degeneracy;
        }
    };

    struct RefComparatorMostBitsFirstLex {

        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy > rhs.degeneracy || ( lhs.degeneracy == rhs.degeneracy && lhs.activeSet.lex_less( rhs.activeSet ) );
        }
    };

    struct RefComparatorLeastBitsFirstLex {

        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy < rhs.degeneracy || ( lhs.degeneracy == rhs.degeneracy && lhs.activeSet.lex_less( rhs.activeSet ) );
        }
    };

    struct RefComparatorClosestToPoint {

        const BitList & mPoint;

        RefComparatorClosestToPoint ( const BitList & point ) :
            mPoint( point ) {}

        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            const unsigned short lhs_dist = mPoint.getCommonPopCnt( lhs.activeSet );
            const unsigned short rhs_dist = mPoint.getCommonPopCnt( rhs.activeSet );

            return lhs_dist < rhs_dist || ( lhs_dist == rhs_dist && mPoint.getInvCommonPopCnt( lhs.activeSet ) < mPoint.getInvCommonPopCnt( rhs.activeSet ) ) ;
        }        
    };

    struct RefComparatorClosestToMultiplePoints {

        const BitList * mPoints;
        const unsigned short mCount;

        RefComparatorClosestToMultiplePoints ( const BitList * points, const unsigned short count ) :
            mPoints( points ),
            mCount( count ) {}

        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            unsigned short lhs_dist;
            unsigned short rhs_dist;

            unsigned short index = 0;

            do {
                lhs_dist = mPoints[ index ].getCommonPopCnt( lhs.activeSet );
                rhs_dist = mPoints[ index ].getCommonPopCnt( rhs.activeSet );

                index++;
            } while ( lhs_dist == rhs_dist && index < mCount );

            return lhs_dist < rhs_dist;
        }        
    };

    struct RefComparatorMaxFrequency {
    
        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.frequency > rhs.frequency;    
        }
    };

    struct RefComparatorMinFrequency {
    
        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.frequency < rhs.frequency;    
        }
    };

    struct RefComparatorMaxSetBitsMaxFrequency {
    
        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy > rhs.degeneracy || ( lhs.degeneracy == rhs.degeneracy && lhs.frequency > rhs.frequency );    
        }
    };

    struct RefComparatorMaxSetBitsMinFrequency {
    
        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy > rhs.degeneracy || ( lhs.degeneracy == rhs.degeneracy && lhs.frequency < rhs.frequency );    
        }
    };

    struct RefComparatorMinSetBitsMinFrequency {
    
        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy < rhs.degeneracy || ( lhs.degeneracy == rhs.degeneracy && lhs.frequency < rhs.frequency );    
        }
    };

    struct RefComparatorMinSetBitsMaxFrequency {
    
        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy < rhs.degeneracy || ( lhs.degeneracy == rhs.degeneracy && lhs.frequency > rhs.frequency );    
        }
    };
}

#endif // __DD_RAY_REFERENCE__
