/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_RAY_REFERENCE_CONTAINER__
#define __DD_RAY_REFERENCE_CONTAINER__

#include <list>
#include <algorithm>

#include "DDTweaks.h"

#include "DDRayReference.h"
#include "DDRayRefVisitor.h"
#include "DDRayRefModifier.h"

namespace double_description {

    class RayReferenceContainer {

        private:
        std::vector<struct RayReference> mRefs;
        BitList mUnionMap;
        BitList mCutMap;

        unsigned int mMinPopCnt;
        unsigned int mMaxPopCnt;
        
        public:
        RayReferenceContainer () :
            mMinPopCnt( UINT_MAX ),
            mMaxPopCnt( 0 ) {}

        RayReferenceContainer ( const RayReferenceContainer & other ) : 
            mRefs( other.mRefs ),
            mUnionMap( other.mUnionMap ),
            mCutMap( other.mCutMap ),
            mMinPopCnt( other.mMinPopCnt ),
            mMaxPopCnt( other.mMaxPopCnt ) {}

        void insert ( const struct RayReference & ref ) {

            if ( unlikely( empty() ) ) {
                mCutMap = ref.activeSet;
            } else {
                mCutMap &= ref.activeSet;
            }

            mRefs.push_back( ref );
            mUnionMap |= ref.activeSet;

            if( mMinPopCnt > ref.degeneracy ) {
                mMinPopCnt = ref.degeneracy;
            }

            if ( mMaxPopCnt < ref.degeneracy ) {
                mMaxPopCnt = ref.degeneracy;
            }
        }

        unsigned int min_pop_cnt() const {
            return mMinPopCnt;
        }

        unsigned int max_pop_cnt() const {
            return mMaxPopCnt;
        }

        unsigned int pop_cnt_range() const {
            return mMaxPopCnt - mMinPopCnt;
        }

        const BitList & union_map () const {
            return mUnionMap;
        }

        const BitList & cut_map () const {
            return mCutMap;
        }       

        unsigned long size () const {
            return mRefs.size();
        }

        bool empty () const {
            return mRefs.empty();
        }

        void rotate () {
            std::swap( mRefs.front(), mRefs.back() );
        }

        const struct RayReference & first () const {
            return mRefs.front();
        }

        const struct RayReference & last () const {
            return mRefs.back();
        }

        const struct RayReference & at ( const unsigned int pos ) const {
            return mRefs.at( pos );
        }

        void iterate ( RayRefVisitor & visitor ) const {  

            do {
                visitor.onStart();

                for ( std::vector<struct RayReference>::const_iterator it_elem = mRefs.begin(); it_elem != mRefs.end(); it_elem++ ) {
                    visitor.onVisit( *it_elem );
                }
        
                visitor.onFinish();
            } while ( visitor.getState() == VS_WAITING );
        }

        void modify ( const RayRefModifier & modifier ) {

            for ( std::vector<struct RayReference>::iterator it_elem = mRefs.begin(); it_elem != mRefs.end(); it_elem++ ) {
                modifier.modify( *it_elem );
            }
        }

       template<class Comparator> void sort ( const Comparator & comp ) {
           std::sort(  mRefs.begin(), mRefs.end(), comp );
        }
       
        void clear () {
            mRefs.clear();
            mUnionMap.clear();
            mCutMap.clear();

            mMinPopCnt = UINT_MAX;
            mMaxPopCnt = 0;
        }
    };
}

#endif // __DD_RAY_REFERENCE_CONTAINER__
